/*----------------------------------------------------------------------*/
/* Foolproof FatFs sample project for AVR              (C)ChaN, 2013    */
/*----------------------------------------------------------------------*/

#include <avr/io.h>	/* Device specific declarations */
#include "ff.h"		/* Declarations of FatFs API */
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/cpufunc.h>
#include "calendar.h"
#include <avr/pgmspace.h>
FATFS FatFs;		/* FatFs work area needed for each volume */
FIL Fil;			/* File object needed for each open file */

#define NUMBER_OF_BOOKS 19833


uint8_t printArray[1024];
volatile uint8_t timerExpire = 0;
volatile uint32_t targetTime = 0;
volatile uint32_t clock = 0;


//the first character in these strings represents the length of the string (not counting the first character)
const uint8_t satSearch[] PROGMEM = "\x3BSearching for satellites. Please place antenna near window.";
const uint8_t satFound[] PROGMEM = "\x2BSatellite time locked. Searching library...";
const uint8_t GPSInitCode[] PROGMEM = "\x33$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28\r\n";
const uint8_t GPSSleepCode[] PROGMEM = "\x0F$PMTK161,0*28\r\n";

const uint8_t fileError1[] PROGMEM = "\x25File Error. Zero Bytes read from file";
const uint8_t fileError2[] PROGMEM = "\x21File Error. Negative book number.";
const uint8_t fileError3[] PROGMEM = "\x21File Error. Book number too high.";
const uint8_t whiteSpace[] PROGMEM = "\x15                     ";
const uint8_t running[] PROGMEM = "\x0aRunning...";
const uint8_t days_in_month[][12] PROGMEM={{31,28,31,30,31,30,31,31,30,31,30,31},{31,29,31,30,31,30,31,31,30,31,30,31}};


void sendbyte(uint8_t byte)
{
	UDR0 = byte;
	while (!(UCSR0A & (0x020)))
		;
}

struct datetime {
    uint8_t day;
    uint8_t month;
    uint16_t year;
    uint8_t hour;
    uint8_t minute;
    uint8_t second;
};


//returns 1 if year is a leap year; else 0
uint8_t isLeapYear(uint16_t year) {
    if (year%4==0) {
        if (year%100==0){
            if (year%400==0) {
                return 1;
            }
        }
        else {
            return 1;
        }
    }
    return 0;
}


uint32_t dayOfCentury(struct datetime * date) {
    int32_t days=0;
    int32_t i;
    for (i=2000;i<date->year;i++) {
        if (isLeapYear(i))
            days+=366;
        else
            days+=365;
    }
    if (isLeapYear(date->year)) {
        for (i=1;i<date->month;i++) {
            days+=pgm_read_byte(&(days_in_month[1][i-1]));
        }
    }
    else {
        for (i=1;i<date->month;i++) {
            days+=pgm_read_byte(&(days_in_month[0][i-1]));
        }
    }
    days+=date->day-1;
    
    return days;
}

uint32_t secondOfDay(struct datetime * date) {
    uint32_t seconds=0;
    seconds+= (uint32_t)date->hour*3600;
    seconds+= (uint32_t)date->minute*60;
    seconds+= (uint32_t)date->second;
    return seconds;
}


//determines how far apart two times are
int32_t secondsApart(struct datetime * firstDate, struct datetime * secondDate) {
    int32_t daysDiff =  (dayOfCentury(secondDate)-dayOfCentury(firstDate));
    int32_t secondsDiff = (secondOfDay(secondDate)-secondOfDay(firstDate));
    return ((daysDiff*86400)+secondsDiff);
    
}



void printString(uint8_t *string, uint32_t length)
{
    sendbyte('~');
    sendbyte('1');
    sendbyte('2');
    sendbyte('8');
    sendbyte('~');
    sendbyte('f');
    sendbyte('0');
    sendbyte('1');
    sendbyte('B');
    sendbyte('\\');
    sendbyte('a');
    sendbyte('\\');
    sendbyte('s');
    sendbyte('\\');
    sendbyte('Y');
    sendbyte('3');
    
    uint32_t i;
    for (i=0;i<length;i++)
    {
        sendbyte(string[i]);
    }
    
    
    
    sendbyte(0x0d);
    sendbyte(0x0d);
    sendbyte(0x0d);
    
}
void sendOpener()
{
    sendbyte('~');
    sendbyte('1');
    sendbyte('2');
    sendbyte('8');
    sendbyte('~');
    sendbyte('f');
    sendbyte('0');
    sendbyte('1');
    sendbyte('B');
    sendbyte('\\');
    sendbyte('a');
    sendbyte('\\');
    sendbyte('s');
    sendbyte('\\');
    sendbyte('Y');
    sendbyte('3');
    
}



void printBits(uint8_t byte)
{
    sendbyte('~');
    sendbyte('1');
    sendbyte('2');
    sendbyte('8');
    sendbyte('~');
    sendbyte('f');
    sendbyte('0');
    sendbyte('1');
    sendbyte('B');
    sendbyte('\\');
    sendbyte('Y');
    sendbyte('3');
    sendbyte('\\');
    sendbyte('d');
    sendbyte('\\');
    sendbyte('s');
    
    
    
    uint32_t i;
    for (i=0;i<8;i++)
    {
        if (((byte<<i)&0x80))
            sendbyte('1');
        else
            sendbyte('0');
    }
    
    
    
    sendbyte(0x0d);
    sendbyte(0x0d);
    sendbyte(0x0d);
    
}

void sendCloser()
{
    sendbyte(0x0d);
    sendbyte(0x0d);
    sendbyte(0x0d);
    
}

void USART_Flush() {
    unsigned char dummy;
    while (UCSR0A & (1<<RXC0)) {
        dummy = UDR0;
    }
}

void sendText(uint8_t *string, uint32_t start, uint32_t end)
{
    uint32_t i;
    for (i=start;i<end;i++)
    {
        sendbyte(string[i]);
    }
}

void sendConstText(const uint8_t *pattern)
{
    uint32_t i;
    for (i=1;i<pgm_read_byte(&(pattern[0]))+1;i++)
    {
        sendbyte(pgm_read_byte(&(pattern[i])));
    }
}
enum LEDStatus {
    on,
    off,
    toggle
};


void setLED (uint8_t  status)
{
    if (status==toggle)
    {
        PORTD ^= (0X08);
    }
    else if (status == on)
        PORTD = PORTD & ~(0X08);
    else
        PORTD = PORTD | (0X08);
}


//receive next serial byte
//return 0 if there's a framing error
inline uint8_t receiveByte()
{
    
    while(!(UCSR0A&0X80))
    {}
    //check for framing error
    if (UCSR0A&0x10) {
        USART_Flush();
        return 0;
    }
    return UDR0;
}

//waits for a specific character to be received over serial
uint8_t waitFor(uint8_t character)
{
    uint8_t receivedByte = receiveByte();
    while (receivedByte != character)
    {
        receivedByte = receiveByte();
        if (receivedByte==0) {
            return 1;
        }
    }
    return 0;
}


uint8_t fillStringBuffer(char *strbuff)
{
    strbuff[0]=receiveByte();
    strbuff[1]=receiveByte();
    if ((strbuff[0]==0)||(strbuff[1]==0)) {
        return 1;
    }
    return 0;
}


//Will attempt to get Satellite 2D or 3D fix
//returns 1 when fixed. Returns 0 when unfixed or error.
uint8_t getFix()
{
    
   uint8_t bufferString[45];
    
    uint8_t i;
    uint8_t start=0;

    USART_Flush();
    uint8_t GPGGA = 0;
    while (GPGGA==0) {
        waitFor('$');
        for (i=0;i<45;i++)
        {
            bufferString[i]=receiveByte();
            
            //if there's been a framing error ABORT
            if (bufferString[i]==0) {
                return 0;
            }
        }
        for (i=0;i<42;i++) {
            if ((bufferString[i]=='G')&&(bufferString[i+1]=='G')&&(bufferString[i+2]=='A')) {
                start=i;
                GPGGA = 1;
                break;
            }
        }
    }
    
    
    uint8_t commaCount = 0;
    for (i=start;i<50&&commaCount<6;i++)
    {
        if (bufferString[i]==',') {
            commaCount++;
        }
    }
    _delay_ms(500);
    if ((bufferString[i]!='0') && (i<50)) {

        return 1;
    }

    return 0;
}

void printNumber(int32_t number)
{
    uint8_t numString[15];
    sprintf((TCHAR*)numString, "%i",number);
    printString(numString,strlen(numString));
    
}


uint8_t getDateTime(struct datetime *currentDateTime)
{

    uint8_t error=0;
    uint8_t loop=1;
    char stringBuffer[2] = "00";
    while (loop) {
        error |= waitFor('$');
        error |= waitFor('G');
        error |= waitFor('P');
        if (receiveByte()=='R') {
            loop=0;
        }
        
    }
    waitFor(',');
    error |= fillStringBuffer(stringBuffer);
    currentDateTime->hour = (uint8_t)(atoi(stringBuffer));
    error |= fillStringBuffer(stringBuffer);
    currentDateTime->minute = (uint8_t)(atoi(stringBuffer));
    error |= fillStringBuffer(stringBuffer);
    currentDateTime->second = (uint8_t)(atoi(stringBuffer));
    uint8_t i;
    for (i=0;i<8;i++) {
        error|=waitFor(',');
    }
    error |= fillStringBuffer(stringBuffer);
    currentDateTime->day = (uint8_t)(atoi(stringBuffer));
    error |= fillStringBuffer(stringBuffer);
    currentDateTime->month = (uint8_t)(atoi(stringBuffer));
    error |= fillStringBuffer(stringBuffer);
    currentDateTime->year = 2000+((uint16_t)(atoi(stringBuffer)));

    return error;
    
}






void ssDown()
{
    PORTB = PORTB & ~(0X04);
}
void ssUp()
{
    PORTB = PORTB | 0X04;
}
void configSD()
{
    
    _delay_ms(5);
    //configure SS MOSI and SCK pins as outputs
    DDRB = DDRB | 0x2C;
    
    
    //Place the SD card in native operating mode
    ssUp(); //ss line up
    PORTB = PORTB | 0x08; //MOSI line up
    uint8_t i;
    for (i=0;i<80;i++)
    {
        PORTB = PORTB | 0x20;
        PORTB = PORTB & ~(0X20);
    }
    
    
    PORTB = PORTB & ~(0X38);
    
    
    
    
    //Config for sample (falling) CPOL=0, CPHA = 0
    SPCR = 0x50; //fosc/4 178kHz
    
    
}



uint8_t SDTransmit(uint8_t data)
{
    
    SPDR = data;
    while(!(SPSR & 0x80))
        ;
    
    return SPDR;
}


void waitForSDClear()
{
    uint8_t returnedByte = 0x00;
    while (returnedByte!=0xff) {
        returnedByte = SDTransmit(0xff);

    }
}
uint8_t sendCommand(uint8_t command,uint8_t arg1, uint8_t arg2, uint8_t arg3, uint8_t arg4, uint8_t crc)
{
    waitForSDClear();
    uint8_t returnedByte = 0xff;
    ssDown();
    SDTransmit(command|0x40);
    SDTransmit(arg1);
    SDTransmit(arg2);
    SDTransmit(arg3);
    SDTransmit(arg4);
    SDTransmit((crc<<1)|1); //CRC
    while(returnedByte==0xff)
        returnedByte=SDTransmit(0xFF);
    
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    ssUp();
    
    return returnedByte;
    
}

void enableRadio()
{
    PORTC = PORTC | 0X02;
}
void disableRadio()
{
    PORTC = PORTC & ~0X02;
}

void printLogFileRow(int32_t targetRow)
{
    uint8_t rowString[38];
    UINT bw;

    f_mount(&FatFs, "", 0);		/* Give a work area to the default drive */
    
    //open another row of the file
    if (f_open(&Fil, "log.txt", FA_READ) == FR_OK) {	/* read file */
        
        f_lseek(&Fil, targetRow*38);
        f_read(&Fil, rowString, 37, &bw);	/* read data from the file */
        
        f_close(&Fil);								/* Close the file */
        
    }
    sendOpener();
    sendText(rowString,0,36);
    sendCloser();
}


void getLogFileRow(uint32_t targetRow,uint8_t* returnString)
{
    UINT bw;
    
    f_mount(&FatFs, "", 0);		/* Give a work area to the default drive */
    
    //open another row of the file
    if (f_open(&Fil, "log.txt", FA_READ) == FR_OK) {	/* read file */
        
        f_lseek(&Fil, targetRow*38);
        f_read(&Fil, returnString, 37, &bw);	/* read data from the file */
        
        f_close(&Fil);								/* Close the file */
        
    }
    
}

void getLogFileRowDate(int32_t targetRow,struct datetime * Date)
{
    UINT bw;
    
    uint8_t logRow[38];
    //open another row of the file
    
    f_lseek(&Fil, targetRow*38);
    f_read(&Fil, logRow, 37, &bw);	/* read data from the file */
        
    
    char yearString[5];
    memcpy(yearString,&logRow[10],4);
    Date->year = atoi(yearString);
    char String[3];
    memcpy(String,&logRow[15],2);
    Date->month = atoi(String);
    memcpy(String,&logRow[18],2);
    Date->day = atoi(String);
    memcpy(String,&logRow[21],2);
    Date->hour = atoi(String);
    memcpy(String,&logRow[24],2);
    Date->minute = atoi(String);
    memcpy(String,&logRow[27],2);
    Date->second = atoi(String);
    
   
    
    
}

void startTimer(uint32_t time)
{
    targetTime+=(time);
    sei();
    timerExpire = 0;
}
void addToTimer(uint32_t time)
{
    targetTime+=time;
}
void waitForTimer()
{
    setLED(on);
    while (timerExpire==0)
    {}
    setLED(off);
    timerExpire = 0;
}




void scrollFile(uint32_t filenumber, uint32_t startChr)
{
    //stores current file name
    TCHAR filepath[20];
    
    UINT bw;
    
    //how much of the file to read this time around
    uint16_t readsize=0;
    
    //generate file name from number
    sprintf( (TCHAR *)filepath, "%u.txt", filenumber);
    
    //size of file
    uint32_t filesize = 0;
    
    //current location
    uint32_t cursor = startChr;
    
    
    //keeps track of accumulating transmission delay
    //always start with opener (16) + closer (3) + spaces after line (16)
    //and add two character's worth of gap because why not?
    uint32_t sizeOfCurrentParagraph = 295;
    
    
    if (f_open(&Fil, filepath, FA_READ) == FR_OK) {	/* read file */
        
        waitForTimer();
        
        //this takes the most time, so the timer needs to be counting while this works.
        filesize = f_size(&Fil);
        f_lseek(&Fil, startChr);
        sendOpener();
        
        //at the end of the file, there won't be 1024 characters left.
        while(cursor<filesize) {
            
            //either read 1024 bytes or just whatever's left in the file
            readsize =(filesize-cursor)<1024?filesize-cursor-1:1024;
            f_read(&Fil, printArray, readsize, &bw);
            
            if (bw==0)
            {
                sendOpener();
                sendConstText(fileError1);
                sendCloser();
                while(1);
                
            }
            
            cursor+=1024;
            

            uint32_t i;
            uint32_t start=0;
            
            //if paragraph is longer than 1024 characters, you need to keep the
            //serial communication open until you reach a return character
            for (i=0;i<readsize;i++)
            {
                if (printArray[i]==13||printArray[i]==10)//carriage return
                {
                    if ((i-start)>2)//don't print lines that are just carriage returns
                    {
                        //add extra space so text can scroll off display safely
                        addToTimer(14);
                        //subtract 1 to account for carriage return
                        startTimer(i-start-1);
                        sizeOfCurrentParagraph+=(i-start-1);
                        sendText(printArray,start,i);
                        sendConstText(whiteSpace);
                        sendCloser();
                        
                        //board can transmit approximately 130 characters in the time it takes the marquee to scroll one character3
                        //the opener and closer account for like 30ms which I consider to be in the marigin of error and don't bother accounting for
                        addToTimer(sizeOfCurrentParagraph/130);
                        sizeOfCurrentParagraph=295;
                        waitForTimer();
                        sendOpener();
                    }
                    start=i;
                }
            }
            addToTimer(readsize-start);
            sizeOfCurrentParagraph+=(readsize-start);
            sendText(printArray,start,readsize);
        }
        //all books are preset to end with carriage return
        //blank the display and add fixed 2.7278 second delay
        //to allow for next file to be read
        sendConstText(whiteSpace);
        sendCloser();
        addToTimer(20);
        f_close(&Fil);								/* Close the file */

    }
}

uint32_t findFileFromDate(struct datetime *currentDateTime, uint32_t *secondsBehind)
{
    f_mount(&FatFs, "", 0);		/* Give a work area to the default drive */
    int32_t currentBook = 0;
    int32_t i = 0;
    struct datetime logDateTime;
    struct datetime nextDateTime;
    if (f_open(&Fil, "log.txt", FA_READ) == FR_OK) {	/* read file */
        
        
        //Get to the beginning of the year in question
        for (i=0;i<NUMBER_OF_BOOKS;i++) {
            getLogFileRowDate(i,&logDateTime);
            
            
            if (logDateTime.year==currentDateTime->year) {
                //year locked
                currentBook = i;
                break;
            }
        }
    

        getLogFileRowDate(currentBook,&logDateTime);
        int32_t difference = secondsApart(&logDateTime,currentDateTime);
        //when difference is less than zero, you're too far ahead and need to go backwards
        
        if (difference<=0) {
            while (difference<=0) {
                
                currentBook--;
                getLogFileRowDate(currentBook,&logDateTime);
                difference = secondsApart(&logDateTime,currentDateTime);
                if (difference>0) {
                    *secondsBehind = difference;
                    break;
                }
                if (currentBook < 0) {
                    sendOpener();
                    sendConstText(fileError2);
                    sendCloser();
                    while(1);
                }
            }
        }
        else {
            while (difference>0) {
                
                currentBook++;
                getLogFileRowDate(currentBook,&logDateTime);
                difference = secondsApart(&logDateTime,currentDateTime);
                if (currentBook > NUMBER_OF_BOOKS) {
                    sendOpener();
                    sendConstText(fileError3);
                    sendCloser();
                    while(1);
                }
            }
            currentBook--; //want to find the book that started before the current time
            getLogFileRowDate(currentBook,&logDateTime);
            *secondsBehind = secondsApart(&logDateTime,currentDateTime);
        }
    }
    
    f_close(&Fil);

    

    return currentBook;
}

uint32_t startBookFromLocation(uint32_t secondsIntoBook, uint32_t filenumber)
{
    if (secondsIntoBook == 0)
        return 0; //save us some time
    
    
    
    //stores current file name
    TCHAR filepath[20];
    
    UINT bw;
    
    //generate file name from number
    sprintf( (TCHAR *)filepath, "%u.txt", filenumber);
    
    //size of file
    uint32_t filesize = 0;
    
    //current location
    uint32_t cursor = 0;
    
    //how far into the book we are
    //measured in "time units" (1 character equivalent)
    uint32_t bookTime = 0;
    
    uint32_t currentLineLength=0;
    
    if (f_open(&Fil, filepath, FA_READ) == FR_OK) {
        
        
        
        filesize = f_size(&Fil);
        
        //calculate number of character time units have elapsed since start of book
        uint32_t bookTimeTarget = secondsIntoBook/.13639;
        
        //at the end of the file, there won't be 1024 characters left.
        while((bookTime<bookTimeTarget)&&(cursor<filesize)) {
            
            f_read(&Fil, printArray, 1024, &bw);
            uint32_t i;
            
            for (i=0;(i<1024)&&(bookTime<bookTimeTarget);i++)
            {
                cursor++;
                if (printArray[i]==10 || printArray[i]==13) {
                    if (currentLineLength >2) {
                        bookTime+=14;
                        bookTime+=((currentLineLength+295)/130);
                    }
                    currentLineLength=0;
                }
                else {
                    currentLineLength++;
                    bookTime+=1;
                }
            }
        }
        f_close(&Fil);
    }
    
    return cursor;
}

int main (void)
{
    setLED(off);
    
    
    //configure GPIO
    DDRD = 0X08; // pin 8 is LED sink
    //config serial
    UCSR0A = 0x00; //no double trans speed
    UCSR0B = 0x18; //RX/TX enable
    UCSR0C = 0x06; //8 bit, asynch, no parity, 1 stop bit
	UBRR0H = 0;
	UBRR0L = 71; //9600 baud @ 11.0592MHz
    
    //When sending messages to the GPS module,
    //everything between $ and * needs to be checksummed.  I used
    //http://www.hhhh.org/wiml/proj/nmeaxor.html
    //turning on RMC and GGA
    //The former is used for time/date, the latter is used to determine level of fix

    
    //configure GPS radio
    _delay_ms(1000);
    _delay_ms(1000);
    sendConstText(GPSInitCode);
    _delay_ms(1000);
    _delay_ms(1000);

    //send UI message to display
    sendOpener();
    sendConstText(satSearch);
    sendCloser();
    
    //Wait for fix
    _delay_ms(1000);
    _delay_ms(1000);
    _delay_ms(1000);
    while (getFix()==0)
    {}
    
    //send UI message
    sendOpener();
    sendConstText(satFound);
    sendCloser();
    _delay_ms(1000);
    

    struct datetime currentDateTime;
    
    

    while (getDateTime(&currentDateTime)==1) {
        _delay_ms(10);
    }
 
    uint32_t secondsIntoBook;
    uint32_t bookDuration;
    uint32_t startFile = findFileFromDate(&currentDateTime, &secondsIntoBook);
    sendConstText(GPSSleepCode);
    uint32_t filenumber;
    //configure timers
    TCCR1A = 0X00;
    TCCR1B = 0X0D; //1024 prescaler
    TIMSK1 = 0x02;//enable timer
    OCR1A = 1472; //should fire once every 136.39ms abouts
    
    
    
    
    
    _delay_ms(1000);
    _delay_ms(1000);
    sendOpener();
    sendConstText(running);
    sendCloser();
    _delay_ms(1000);
    _delay_ms(1000);
    _delay_ms(1000);
    
    f_mount(&FatFs, "", 0);		/* Give a work area to the default drive */
    
    timerExpire = 1;
    
    while(1)
    {
        filenumber = startFile;
        while (filenumber<NUMBER_OF_BOOKS) {
            scrollFile(filenumber, startBookFromLocation(secondsIntoBook,filenumber));
            filenumber++;
            secondsIntoBook = 0;
        }
        startFile = 0;
    }
    for(;;);
}



/*---------------------------------------------------------*/
/* User Provided RTC Function called by FatFs module       */

DWORD get_fattime (void)
{
	/* Returns current time packed into a DWORD variable */
	return	  ((DWORD)(2013 - 1980) << 25)	/* Year 2013 */
			| ((DWORD)7 << 21)				/* Month 7 */
			| ((DWORD)28 << 16)				/* Mday 28 */
			| ((DWORD)0 << 11)				/* Hour 0 */
			| ((DWORD)0 << 5)				/* Min 0 */
			| ((DWORD)0 >> 1);				/* Sec 0 */
}

ISR(TIMER1_COMPA_vect)
{

    clock++;
    if (clock==targetTime)
        timerExpire = 1;
}

