//Firmware for Gutenberg Clock

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>


uint8_t printArray[256];

void sendbyte(uint8_t byte)
{
	UDR0 = byte;
	while (!(UCSR0A & (0x020)))
		;
}


void printString(uint8_t *string, uint32_t length)
{
    sendbyte('~');
    sendbyte('1');
    sendbyte('2');
    sendbyte('8');
    sendbyte('~');
    sendbyte('f');
    sendbyte('0');
    sendbyte('1');
    sendbyte('B');
    sendbyte('\\');
    sendbyte('a');
    sendbyte('\\');
    sendbyte('s');
    sendbyte('\\');
    sendbyte('Y');
    sendbyte('3');
    
    uint32_t i;
    for (i=0;i<length;i++)
    {
        sendbyte(string[i]);
    }
    
    
    
    sendbyte(0x0d);
    sendbyte(0x0d);
    sendbyte(0x0d);
    
}



enum LEDStatus {
    on,
    off
};


void setLED (uint8_t  status)
{
    if (status == on)
        PORTD = PORTD & ~(0X08);
    else
        PORTD = PORTD | (0X08);
}

void ssDown()
{
    PORTB = PORTB & ~(0X04);
}
void ssUp()
{
    PORTB = PORTB | 0X04;
}
void configSD()
{
    
    _delay_ms(5);
    //configure SS MOSI and SCK pins as outputs
    DDRB = DDRB | 0x2C;
    
    
    //Place the SD card in native operating mode
    ssUp(); //ss line up
    PORTB = PORTB | 0x08; //MOSI line up
    uint8_t i;
    for (i=0;i<80;i++)
    {
        PORTB = PORTB | 0x20;
        PORTB = PORTB & ~(0X20);
    }

    
    PORTB = PORTB & ~(0X38);
    
    
    
    
    //Config for sample (falling) CPOL=0, CPHA = 0
    SPCR = 0x50; //fosc/4 178kHz

    
}



uint8_t SDTransmit(uint8_t data)
{

    SPDR = data;
    while(!(SPSR & 0x80))
        ;

    return SPDR;
}

void printBits(uint8_t byte)
{
    sendbyte('~');
    sendbyte('1');
    sendbyte('2');
    sendbyte('8');
    sendbyte('~');
    sendbyte('f');
    sendbyte('0');
    sendbyte('1');
    sendbyte('B');
    sendbyte('\\');
    sendbyte('a');
    sendbyte('\\');
    sendbyte('s');
    sendbyte('\\');
    sendbyte('Y');
    sendbyte('3');
    
    uint32_t i;
    for (i=0;i<8;i++)
    {
        if (((byte<<i)&0x80))
            sendbyte('1');
        else
            sendbyte('0');
    }
    
    
    
    sendbyte(0x0d);
    sendbyte(0x0d);
    sendbyte(0x0d);

}
void waitForSDClear()
{
    uint8_t returnedByte = 0x00;
    while (returnedByte!=0xff)
        returnedByte = SDTransmit(0xff);
}
uint8_t sendCommand(uint8_t command,uint8_t arg1, uint8_t arg2, uint8_t arg3, uint8_t arg4, uint8_t crc)
{
    waitForSDClear();
    uint8_t returnedByte = 0xff;
    ssDown();
    SDTransmit(command|0x40);
    SDTransmit(arg1);
    SDTransmit(arg2);
    SDTransmit(arg3);
    SDTransmit(arg4);
    SDTransmit((crc<<1)|1); //CRC
    while(returnedByte==0xff)
        returnedByte=SDTransmit(0xFF);
    
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    SDTransmit(0xff);
    ssUp();

    return returnedByte;
    
}



int main(void)
{
    DDRD = 0X18;

    //config serial
    UCSR0A = 0x00; //no double trans speed
    UCSR0B = 0x08; //TX enable
    UCSR0C = 0x46; //8 bit, asynch, no parity, 1 stop bit
	UBRR0H = 0;
	UBRR0L = 71; //9600 baud
    
    
    
    configSD();
    uint8_t i;
    uint8_t returnedByte = sendCommand(0,0,0,0,0,0x4a); //need to send CMD0 with valid CRC
    
    //Then wait for idle response
    while (returnedByte!=0x01) {
        ssDown();
        returnedByte = SDTransmit(0xff);
        ssUp();
    }
    
    //for SDHC cards, you need to send CMD8 once to enable the SDHC set bit in ACMD41
    returnedByte = sendCommand(0x48,0,0,1,0x55,0x3a); //set to 2.7-3.6V; check pattern set to 0x55
    
    while (returnedByte==1)
    {
        sendCommand(55,0,0,0,0,0x32);
        returnedByte = sendCommand(41,0x40,0x00,0x00,0x00,0x7B); //need to set High capacity bit (bit 30) to 1
    }

    
    
    
    for (i=91;i<250;i++)
    {
        waitForSDClear();
        ssDown();
        SDTransmit(17|0x40);
        SDTransmit(0x00);
        SDTransmit(0x00);
        SDTransmit(i);
        SDTransmit(0x00);
        SDTransmit(0x95); //CRC
        
        //wait for Data Token
        returnedByte= 0x00;
        while(returnedByte !=0xFE)
            returnedByte = SDTransmit(0xFF);
        
        //get data
        uint32_t blockCount = 0;
        for (blockCount=0;blockCount<256;blockCount++) {
            printArray[blockCount]=SDTransmit(0xFF);
        }
        
        for (blockCount=0;blockCount<256;blockCount++) {
            SDTransmit(0xFF);
        }
        
        //CRC
        SDTransmit(0xFF);
        SDTransmit(0xFF);
        
        ssUp();
        _delay_ms(1);
        printString(printArray,256);
        while(1);
    }
   
    
    
    uint8_t myphrase[11]={"HELLO-WORLD"};
    printString(myphrase,11);

    while(1)
    {
    }
    
}